#!/usr/bin/perl
#
# this script converts Sokoban maps into a c data structure
# run like: ./convert.pl < Sokoban.def > levels.h
#

$y = 0;
$width = 22;
$height = 16;
$level = 0;
@line = ();
$longest = 22;

%values =
  (
   '*' => 0,
   '#' => 1,
   ' ' => 2,
   '$' => 3,
   '.' => 4,
   '@' => 5,
   'O' => 6,
  );

print "const char levelList[85][$width*$height] = {\n";
while (<>) {
  chomp;
  if(m/;LEVEL (\d+)/) { # new level, print old
    printlevel($1-1);
  } elsif (!/^;/) { # not a comment
    push @line, $_;
    if (length > $longest) {
      $longest = length;
    }
  }
}

# at EOF, print last level
printlevel("last");

print "};\n";

sub printlevel {

  $level = shift;

  if($level) { # avoid printing empty level at start

      print "\n    // this is level $level\n\n";

      # centre the level
      foreach $line (@line) {
	  $line = (" " x (int(($width-$longest)/2)) . $line);
      }
      for ($i = 1; $i<($height - $#line)/2; $i++) {
	  unshift @line, "";
      }
      print "  {\n"; # new level starting
      
      foreach (@line) {
	  print "    ";
	  $y++;
	  for ($x = 0; $x < $width; $x++) {
	      if ($c = substr($_,$x,1)) {
		  print("$values{$c}, ");
	      } else {
		  print "2, ";
	      }
	  }
	  print "\n";
      }
      for ($y; $y<$height; $y++) {
	  print "    " . ("2, " x $width) . "\n";
      }
      print "  },\n\n";
      
      @line = ();
      $y = 0;
      $longest = 0;
  
  }
}
