#include <pspkernel.h>
#include <pspdebug.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include "pg.h"

/* Define the module info section */
PSP_MODULE_INFO("Sokoban", 0, 1, 1);

/* Button bit masks */ 
#define CTRL_SQUARE     0x8000 
#define CTRL_TRIANGLE   0x1000 
#define CTRL_CIRCLE     0x2000 
#define CTRL_CROSS      0x4000 
#define CTRL_UP         0x0010 
#define CTRL_DOWN       0x0040 
#define CTRL_LEFT       0x0080 
#define CTRL_RIGHT      0x0020 
#define CTRL_START      0x0008 
#define CTRL_SELECT     0x0001 
#define CTRL_LTRIGGER   0x0100 
#define CTRL_RTRIGGER   0x0200 

#define VERSION "0.9"

#define PIXELSIZE 1
#define LINESIZE 512

#define RGB(r, g, b) ((r)|((g)<<5)|((b)<<10))

#define WIDTH 22
#define HEIGHT 16

#define LEVELS 85

#define GOALANDBOX 0
#define WALL 1
#define FLOOR 2
#define BOX 3
#define GOAL 4
#define PLAYER 5
#define PLAYER_N 0
#define PLAYER_E 1
#define PLAYER_W 2
#define PLAYER_S 3

#define BACKGROUND_COLOR RGB(0, 0, 0)
#define HELP_TEXT RGB(0, 0, 0)
#define HELP_BORDER RGB(0x00,0x00,0x00)
#define HELP_COLOR RGB(0xff, 0xff, 0xff)

#include "tiles.h"
#include "levels.h"

unsigned short int highscores[LEVELS+1];

/*
  lots of icky global variables follow
*/

char gameBoard[WIDTH*HEIGHT];  // the current level
char levelGoals[WIDTH*HEIGHT]; // where the levelGoals for boxes are. bad design but works

/*
  undolistEnd holds the spot a box was pushed to or the player moved to
  16k moves should be enough for any level, memory is cheap!
  else I could've used a linked list or so
*/
short int undoTarget[16384];
// holds the heading
char undoHeading[16384];

int currentLevel = 1;// fairly obvious!
int playerPos;       // the player's current position as offset in gameBoard
int playerHeading=0; // the players current heading as an offset from PLAYER_N in tileSet
int moveCount=0;     // number of moves on current level
int wonYet=0;        // has the player won yet, the slacker?


int exit_callback(void)
{
	sceKernelExitGame();
	return 0;
}

int CallbackThread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}


void createScreenshot() {

  unsigned char  bmpHeader24[] = { 0x42, 0x4d, 0x38, 0xfa, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x36, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, 0x00, 0xe0, 0x01, 0x00, 0x00, 0x10, 0x01, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x12, 0x0b, 0x00, 0x00, 0x12, 0x0b, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
  unsigned char  buffer[SCREEN_WIDTH*3];
  unsigned char  r,g,b;
  int bufferIndex = 0;
  unsigned short p;
  int file;
  unsigned char *vptr;
  unsigned char *vptr0;
  int i=0;
  int e=0;
  char savePath[] = "ms0:/ss.bmp";

  file = sceIoOpen(savePath,O_CREAT | O_TRUNC | O_RDWR, 0777);

  // write bmp header
  sceIoWrite(file,bmpHeader24,54);
   
  // write bmp data
  vptr0 = pgGetVramAddr(0,271);   // you can find this function in NEM's sources

  for(i=0; i<SCREEN_HEIGHT; i++)
    {
      //printDecimal(0, i%10, RGB(0xff,0xff,0xff), vptr0);  
      vptr=vptr0;
      for(e=0; e<SCREEN_WIDTH; e++)
	{
	  p = *(unsigned short *)vptr;
	  r = (p<<3)&0xf8;
	  g = (p>>2)&0xf8;
	  b = (p>>7)&0xf8;         
         
	  buffer[bufferIndex] = b;
	  bufferIndex++;         
	  buffer[bufferIndex] = g;
	  bufferIndex++;         
	  buffer[bufferIndex] = r;
	  bufferIndex++;

	  vptr+=PIXELSIZE*2;
	}
      // write chunk
      sceIoWrite(file,buffer,SCREEN_WIDTH*3);
      bufferIndex=0;
      vptr0-=LINESIZE*2;
    }

  // bmp end
  unsigned char end[] = { 0x00, 0x00 };
  sceIoWrite(file,end,2);

  sceIoClose(file);

}

void printDecimal(int x, int y, int color, int value) {
  int i, count=0;
  char digits[6];
  for (i = 5; i >= 0; i--) {
    digits[i] = '0' + (value%10);
    value /= 10;
  }
  i = 0;
  while (i < 5 && digits[i] == '0') {
    digits[i++] = 0;
  }
  for (i = 0; i < 6; i++) {
    if (digits[i]) {
      count++;
      pgPutChar((x++)*8, y*8, color, BACKGROUND_COLOR, digits[i], 1, 1, 1);
    }
  }
  while (count < 6) {
      pgPutChar((x++)*8, y*8, color, BACKGROUND_COLOR, ' ', 1, 1, 1);
      count++;
  }
}


void fillCell(int x, int y, int color)
{
  unsigned short* vram = (unsigned short*) pgGetVramAddr(x * 16, y * 16);
  int xo, yo;
  for (xo = 0; xo < 16; xo++) {
    for (yo = 0; yo < 16; yo++) {
      vram[xo + yo * 512] = color;
    }
  }
}


void drawTile(int pos) {
  // transparency is an ugly hack because my png2c doesn't support png transparency

  // need to draw a floor tile or goal beneath player or box
  if (gameBoard[pos] == PLAYER || gameBoard[pos] == BOX) {
    if (levelGoals[pos]) {
      pgBitBltTransp( ((pos%WIDTH)+8)*16, (pos/WIDTH)*16+8, 16, 16, 1, tileSet[GOAL],0x354c);
    } else {
      pgBitBltTransp( ((pos%WIDTH)+8)*16, (pos/WIDTH)*16+8, 16, 16, 1, tileSet[FLOOR],0x354c);
    }
  }
  if(gameBoard[pos] == PLAYER) {
    pgBitBltTransp( ((pos%WIDTH)+8)*16, (pos/WIDTH)*16+8, 16, 16, 1, tileSet[gameBoard[pos]+playerHeading],0x354c);
  } else {
    pgBitBltTransp( ((pos%WIDTH)+8)*16, (pos/WIDTH)*16+8, 16, 16, 1, tileSet[gameBoard[pos]],0x354c);
  }
}

int checkWin() {
  // check if all goals have boxes on them, if so, player wins
  int pos;
  for (pos = 0; pos < WIDTH*HEIGHT; pos++) {
    if (levelGoals[pos] && gameBoard[pos] != BOX) {
      return 0;
    }
  }
  return 1;
}

void move(int dx, int dy) {
  int target;
  int next;

  if(dx) {
    if(dx>0) {
      playerHeading = PLAYER_E;
    } else {
      playerHeading = PLAYER_W;
    }
    target = playerPos+dx;
    next = playerPos+dx*2;
  } else {
    if(dy>0) {
      playerHeading = PLAYER_S;
    } else {
      playerHeading = PLAYER_N;
    }
    target = playerPos+dy*WIDTH;
    next = playerPos+dy*WIDTH*2;
  }

  // save heading for undo
  undoHeading[moveCount] = playerHeading;

  if (gameBoard[target] == FLOOR || gameBoard[target] == GOAL) {

    // move ends at target, save it for undo
    undoTarget[moveCount] = target;

    gameBoard[target] = PLAYER;
    drawTile(target);

    gameBoard[playerPos] = levelGoals[playerPos] ? GOAL : FLOOR;
    drawTile(playerPos);

    playerPos = target;

    moveCount++;
    printDecimal(7, 7, RGB(0xff,0x00,0x00), moveCount);


  } else if (gameBoard[target] == BOX && (gameBoard[next] == FLOOR || gameBoard[next] == GOAL)) {

    // move ends at next, save it for undo
    undoTarget[moveCount] = next;

    gameBoard[next] = BOX;
    drawTile(next);

    gameBoard[target] = levelGoals[target] ? GOAL : FLOOR;
    drawTile(target);

    gameBoard[target] = PLAYER;
    drawTile(target);

    gameBoard[playerPos] = levelGoals[playerPos] ? GOAL : FLOOR;
    drawTile(playerPos);

    playerPos = target;

    moveCount++;
    printDecimal(7, 7, RGB(0xff,0x00,0x00), moveCount);

    if(levelGoals[next]) {
      // we pushed a box onto a goal, did we win?
      wonYet = checkWin();
      // (ugly global var checked by main loop)
    }
  }
}

void undo() {
  // Works like move() in reverse, i.e. boxes push the player
  if (moveCount) { // can't go negative!
    moveCount--;
    if (moveCount) {
      // change heading to the previous one, unless back at start
      playerHeading = undoHeading[moveCount-1];
    } else {
      playerHeading = PLAYER_N;
    }

    int target;
    int next;

    if (undoHeading[moveCount] == PLAYER_E) {
      target = undoTarget[moveCount]-1;
      next = undoTarget[moveCount]-2;
    } else if (undoHeading[moveCount] == PLAYER_W) {
      target = undoTarget[moveCount]+1;
      next = undoTarget[moveCount]+2;
    } else if (undoHeading[moveCount] == PLAYER_S) {
      target = undoTarget[moveCount]-WIDTH;
      next = undoTarget[moveCount]-2*WIDTH;
    } else {
      target = undoTarget[moveCount]+WIDTH;
      next = undoTarget[moveCount]+2*WIDTH;
    }

    // we don't need to check if a move is valid here since move() did that

    if (gameBoard[undoTarget[moveCount]] == PLAYER) {
      // didn't push a box, just move the player back to target
      gameBoard[target] = PLAYER;
      drawTile(target);
      
      playerPos = target;

    } else if ( gameBoard[undoTarget[moveCount]] == BOX ) {

      // need to draw backgrounds beneath player/box also
      gameBoard[next] = levelGoals[target] ? GOAL : FLOOR;
      drawTile(next);
      gameBoard[next] = PLAYER;
      drawTile(next);
      
      gameBoard[target] = levelGoals[target] ? GOAL : FLOOR;
      drawTile(target);
      gameBoard[target] = BOX;
      drawTile(target);

      playerPos = next;
    }
    // in either case, draw the vacated tile
    gameBoard[undoTarget[moveCount]] = levelGoals[undoTarget[moveCount]] ? GOAL : FLOOR;
    drawTile(undoTarget[moveCount]);

    printDecimal(7, 7, RGB(0xff,0x00,0x00), moveCount);
    //drawBoard(0);

  }
}

char* highscoreFilename() {
  char filename[255];
  int file;
  /*
    path is in g_elf_name without the last 9 chars = "eboot.pbp"
    buffer overflow possibilities galore, but who cares? 
    just say no to ridiculously long path and we're fine
  */
  strncpy(filename, g_elf_name, strlen(g_elf_name)-9); 
  filename[strlen(g_elf_name)-9]=0;
  strcat(filename, "scores.dat");
  return filename;
}

void loadHighscores() {
  char* filename;
  filename = highscoreFilename();
  int file;

  if (0 > (file = sceIoOpen(filename, PSP_O_RDONLY, 0777))) {

    // there is no highscore file yet
    // initialise highscores
    int i;
    for (i = 1; i <= LEVELS; i++) {
      highscores[i] = 0;
    }

  } else {

    sceIoRead(file, highscores, (LEVELS+1)*2);
    sceIoClose(file);

  }

}

void saveHighscores() {
  char* filename;
  filename = highscoreFilename();
  int file;

  // open/create highscore file
  file = sceIoOpen(filename,O_CREAT | O_TRUNC | O_RDWR, 0777);

  // write scores to file
  sceIoWrite(file,highscores,(LEVELS+1)*2);
  sceIoClose(file);

}


void drawBoard(int load)
{
  pgFlipDrawFrame();
  pgFillvram(BACKGROUND_COLOR);

  int pos;
  for (pos = 0; pos < WIDTH*HEIGHT; pos++) {
    if (load) { 
      gameBoard[pos] = levelList[currentLevel-1][pos];
    }
    if (gameBoard[pos] == GOALANDBOX) {
      levelGoals[pos] = 1;
      gameBoard[pos] = BOX;
    } else if (gameBoard[pos] == GOAL) {
      levelGoals[pos] = 1;
    } else {
      levelGoals[pos] = 0;
    }
    if (gameBoard[pos] == PLAYER) { // player position
      playerPos = pos;
    }
    drawTile(pos);
  }
  pgPrint2(0, 1, RGB(0xff,0x00,0x00), "Sokoban");
  
  pgPrint(0, 5, RGB(0xff,0x00,0x00), "Level:");
  printDecimal(7, 5, RGB(0xff,0x00,0x00), currentLevel);
  
  pgPrint(0, 7, RGB(0xff,0x00,0x00), "Moves:");
  printDecimal(7, 7, RGB(0xff,0x00,0x00), moveCount);

  if (highscores[currentLevel]) {
    pgPrint(0, 9, RGB(0xff,0x00,0x00),  "Record:");
    printDecimal(8, 9, RGB(0xff,0x00,0x00), highscores[currentLevel]);
  } else {
    pgPrint(0, 9, RGB(0xff,0x00,0x00),  "Unsolved level");
  }

  pgPrint(1, 30, RGB(0xff,0x00,0x00), "Press [start]");
  pgPrint(1, 31, RGB(0xff,0x00,0x00), "for help");

  pgFlipShowFrameV();
  
}

void newGame() {
  wonYet = 0;
  moveCount=0;
  drawBoard(1);
}

void drawBox(int x, int y, int w, int h, int border, int color) {
  unsigned short *vram = (unsigned short*) pgGetVramAddr(x * 16, y * 16);
  unsigned short p;
  int xo, yo;
  for (xo = 0; xo <= 16*w; xo++) {
    for (yo = 0; yo <= 16*h; yo++) {
      if (xo == 0 || xo == 16*w || yo == 0 || yo == 16*h) {
	vram[xo + yo * 512] =  border;
      } else {
	p = vram[xo + yo * 512];

	vram[xo + yo * 512] =  color;
      }
    }
  }
}

void showHelp() {

  drawBox(5,2,20,9, HELP_BORDER, HELP_COLOR);

  pgPrint(14, 7, HELP_TEXT, "PSP Sokoban " VERSION);

  pgPrint(14, 9, HELP_TEXT,  "How to play: Push all the objects");
  pgPrint(14, 10, HELP_TEXT, "into the goals. You can only push");
  pgPrint(14, 11, HELP_TEXT, "one object at a time.");

  pgPrint(14, 13, HELP_TEXT, " Controls: Move with D-Pad.");
  pgPrint(14, 14, HELP_TEXT, " [Select] will undo a move.");
  pgPrint(14, 15, HELP_TEXT, " Right trigger for next level.");
  pgPrint(14, 16, HELP_TEXT, " Left trigger for previous.");

  pgPrint(14, 18, HELP_TEXT, "Press [start] to continue");


  pgWaitVn(20);
  ctrl_data_t ctrl;
  while (1) {
    pgWaitV();
    sceCtrlReadBufferPositive(&ctrl, 1);
    if (ctrl.buttons & CTRL_START) {
      break;
    }
  }
  drawBoard(0);
  pgWaitVn(20);
}

void keyboardControl()
{
  ctrl_data_t ctrl;
  sceCtrlReadBufferPositive(&ctrl, 1); 
  int dx=0, dy=0, dpad=0;
  if (ctrl.buttons & CTRL_UP) { 
    dpad = 1;
    dx = 0;
    dy = -1;
  } else if (ctrl.buttons & CTRL_DOWN) { 
    dpad = 1;
    dx = 0;
    dy = 1;
  } else if (ctrl.buttons & CTRL_LEFT) { 
    dpad = 1;
    dx = -1;
    dy = 0;
  } else if (ctrl.buttons & CTRL_RIGHT) { 
    dpad = 1;
    dx = 1;
    dy = 0;
  } else if (ctrl.buttons & CTRL_LTRIGGER) {
    if ( currentLevel > 1 ) {
      currentLevel--;
    }
    pgWaitVn(10);
    newGame();
  } else if (ctrl.buttons & CTRL_RTRIGGER) {
    if ( currentLevel < LEVELS ) {
      currentLevel++;
      pgWaitVn(10);
    }
    pgWaitVn(10);
    newGame();
  } else if (ctrl.buttons & CTRL_TRIANGLE) {
    // make a screenshot
    createScreenshot();
  } else if (ctrl.buttons & CTRL_START) {
    // show help message
    showHelp();
  } else if (ctrl.buttons & CTRL_SELECT) {
    // undo
    undo();
    pgWaitVn(10);
  }



  if (dpad && !wonYet) {
    move(dx,dy);
    pgWaitVn(10);
  }
}


int main() {
  SetupCallbacks();

  pgInit();
  pgScreenFrame(1, 1);

  loadHighscores();

  newGame();
	
  while (1) {
    keyboardControl();
    if (wonYet) {

      if (highscores[currentLevel] == 0 || moveCount<highscores[currentLevel]) {
	highscores[currentLevel] = moveCount;
	drawBoard(0); // lazy way of updating record display
	saveHighscores();
      }

      drawBox(7,5,14,5, HELP_BORDER, HELP_COLOR);
      pgPrint2(10, 6, RGB(0x00,0x00,0xff), "You Win!");
      pgPrint(18, 17, RGB(0x00,0x00,0x00), "Press right trigger");
      pgPrint(18, 18, RGB(0x00,0x00,0x00), "for the next level");

    }
    pgWaitVn(1);
    /*
      printDecimal(47, 12, TEXT_COLOR, score);
      printDecimal(47, 18, TEXT_COLOR, high);
    */
  }
  return 0;
}


